import json

import shutil


class DbQueryService:
    def __init__(self, app, default_headers, options):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__default_headers = default_headers
        self.__options = options

    def schema_data(self, configuration):
        params = {"configuration": json.dumps(configuration)}
        dr = self.__app.native_api_call('db', 'schema-data', params, self.__options, True)
        return json.loads(dr.text)

    def upload_data(self, file_descriptor, configuration):
        multipart_form_data = {
            'file': file_descriptor
        }
        params = {"configuration": json.dumps(configuration)}
        dr = self.__app.native_api_call('db', 'upload-data', params, self.__options, True, multipart_form_data)
        return json.loads(dr.text)

    def download_data(self, configuration, output_file):
        params = {"configuration": json.dumps(configuration)}
        response = self.__app.native_api_call('db', 'download-data', params, self.__options, True, None, True)
        with open(output_file, 'wb') as out_file:
            shutil.copyfileobj(response.raw, out_file)
        del response

    def batch_update(self, command, rows):
        """
        Для массовой вставки умеренных объемов 1-5к записей за вызов

        :param command: SQL insert or updtae
        :param rows: list of dict
        :return: dict
        """
        request = {
            "database": {
                "alias": self.__options['dbAlias']
            },
            "batchUpdate": {
                "command": command,
                "rows": rows,
                "shardKey": self.__options.get('shardKey'),
            }
        }
        dr = self.__app.native_api_call('db', 'batch-update', request, self.__options, False)
        return json.loads(dr.text)

    def update(self, command, params=None):
        """
        Запросы на INSERT, UPDATE, DELETE и пр. не возвращающие результата должны выполняться через этот метод
        Исключение такие запросы с RETURNING для PostgreSQL

        :param command: SQL запрос
        :param params: Параметры для prepared statements
        :rtype: object DataResult
        """
        request = {
            "database": {
                "alias": self.__options['dbAlias']
            },
            "dbQuery": {
                "command": command,
                "parameters": params,
                "shardKey": self.__options.get('shardKey'),
            }
        }
        dr = self.__app.native_api_call('db', 'update', request, self.__options, False)
        return json.loads(dr.text)

    def query(self, command, params=None, max_rows=0):
        """
        Выполняет запрос, который ОБЯЗАТЕЛЬНО должен вернуть результат.
        Если вам надо сделать INSERT, UPDATE, DELETE или пр. используйте метод update
        или возвращайте результата через конструкцию RETURNING (нет в MySQL)

        > db.query('SELECT * FORM users WHERE id=:id', {"id":MY_USER_ID})

        :param command: SQL запрос
        :param params: Параметры для prepared statements
        :param max_rows: Если запрос вернет строк больше, чем указано в max_rows, то будет ошибка. Если равно нулю, действуют стандартные ограничения (50000)
        :rtype: object DataResult
        """
        request = {
            "database": {
                "alias": self.__options['dbAlias']
            },
            "dbQuery": {
                "maxRows": max_rows,
                "command": command,
                "parameters": params,
                "shardKey": self.__options.get('shardKey'),
            }
        }
        dr = self.__app.native_api_call('db', 'query', request, self.__options, False)
        return json.loads(dr.text)

    def one(self, command, params=None):
        """
        Возвращает первую строку ответа, полученного через query

        > db.query('SELECT * FORM users WHERE id=:id', {"id":MY_USER_ID})

        :param command: SQL запрос
        :param params: Параметры для prepared statements
        :rtype: dict
        """
        dr = self.query(command, params)
        if dr['rows']:
            return dr['rows'][0]
        else:
            return None

    def all(self, command, params=None):
        """
        Возвращает строки ответа, полученного через query

        > db.query('SELECT * FORM users WHERE id=:id', {"id":MY_USER_ID})

        :param command: SQL запрос
        :param params: Параметры для prepared statements
        :rtype: list of dict
        """
        dr = self.query(command, params)
        return dr['rows']



import re

import argparse

import docker

from metasdk import read_developer_settings


def generate_tag(service: str) -> str:
    return "{}-latest".format(service)


def get_build_num():
    import datetime
    nd = datetime.datetime.now()
    fd = datetime.datetime.strftime(nd, "%Y%m%d%H%M%S")
    return fd


def await_build(build_result, image_tag):
    response = []
    srch = 'Successfully tagged ' + image_tag
    for line in build_result:
        line_get = line.get('stream', '')
        print("Docker build: " + line_get.replace("\n", ""))
        response.append(line_get)

        match = re.search(srch, line_get)
        if match:
            print("Build OK", {"res": response[-10:]})
            return
    raise Exception("Could not build image. {}".format(response[-10:]))


def await_push(push_result, service_docker_tag):
    response = []
    srch = r'{}: digest: sha256:[0-9a-f]+'.format(service_docker_tag)
    for line in push_result:
        response.append(line.get('stream', ''))
        match = re.search(srch, line.get('status', ''))
        if match:
            print("Push result", {"res": response[-10:]})
            return
    raise Exception("Could not push image. {}".format(response[-10:]))


def build(grpc_service, docker_repo, workdir):
    gcloud_params = read_developer_settings().get('gcloudDev')
    if not gcloud_params:
        raise ValueError("gcloudDev не установлены в developer_settings")
    gcloud_prefix = gcloud_params.get('prefix', '')
    endpoint_service_name = gcloud_prefix + "-" + grpc_service if gcloud_prefix else grpc_service
    service_docker_tag = endpoint_service_name + "-latest"

    docker_cli = docker.from_env()
    build_and_push(docker_repo, service_docker_tag, workdir, docker_cli)


def build_and_push(docker_repo: str, service_docker_tag: str, workdir: str, docker_cli):
    image_tag = "{}:{}".format(docker_repo, service_docker_tag)
    print("Build start")
    await_build(
        docker_cli.api.build(
            path=workdir,
            tag=image_tag,
            rm=True,
            stream=True,
            decode=True
        )
        , image_tag)
    print("Push start")
    await_push(
        docker_cli.api.push(
            repository=docker_repo,
            tag=service_docker_tag,
            stream=True,
            decode=True
        ),
        service_docker_tag
    )


def main():
    """
    Билд и деплой docker image
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--service', help='Name of API Service. Example: hello', type=str, required=True)
    parser.add_argument('--docker_repo', help='Name of docker repo. Example: mycompany/repo', type=str, required=True)
    parser.add_argument('--workdir', help='Root of project dir. Default "."', default=".", type=str, required=False)
    args = parser.parse_args()

    build(args.service, args.docker_repo, args.workdir)


if __name__ == '__main__':
    main()



import os
import subprocess


def exec_cmd(cmd, check=True):
    print("Rin cmd:")
    print(cmd)
    subprocess.run(cmd, shell=True, check=check)


os.chdir(os.path.dirname(os.path.abspath(__file__)))
TOOLS_DIR = os.getcwd()



import argparse
from os.path import expanduser

from metasdk import read_developer_settings
from metasdk.internal import OS_NAME
from metasdk.tools import exec_cmd


def main():
    """
    Запускает локальный Docker контейнер с google cloud esp
    Это нужно для транскодирования REST запросов на ваш GRPC backend, а так же проверки авторизации
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--service', help='Name of API Service. Example: hello', type=str, required=True)
    parser.add_argument('--backend_port', help='Backend grpc app port. Default: 50051', type=int, required=False, default=50051)
    parser.add_argument('--frontend_http_port', help='ESP Frontend port for HTTP(REST) requests. Default: 8083', type=int, required=False, default=8083)
    parser.add_argument('--frontend_grpc_port', help='ESP Frontend port for GRPC request. Default: 8084', type=int, required=False, default=8084)
    parser.add_argument('--service_account_key_file', help='Service account json file for trace and log google cloud api. Default: test-esp-service-account-creds.json', type=str,
                        required=False, default="test-esp-service-account-creds.json")
    args = parser.parse_args()

    grpc_service = args.service
    frontend_http_port = args.frontend_http_port
    backend_port = args.backend_port
    frontend_grpc_port = args.frontend_grpc_port

    run_serve(grpc_service, backend_port, frontend_http_port, frontend_grpc_port)


def run_serve(grpc_service, backend_port=50051, frontend_http_port=8083, frontend_grpc_port=8084):
    gcloud_params = read_developer_settings().get('gcloudDev')
    if not gcloud_params:
        raise ValueError("gcloudDev не установлены в developer_settings")
    gcloud_project = gcloud_params['project']
    gcloud_prefix = gcloud_params.get('prefix', '')
    endpoint_service_name = gcloud_prefix + "-" + grpc_service if gcloud_prefix else grpc_service
    user_dir = expanduser("~")
    if OS_NAME == "macos":
        backend_host = "docker.for.mac.localhost"
    elif OS_NAME == "windows":
        backend_host = "docker.for.win.localhost"
    else:
        backend_host = "localhost"
    print("Start proxy server on: http://localhost:" + str(frontend_http_port))
    exec_cmd("""
        docker run \
            --rm \
            --publish={frontend_http_port}:{frontend_http_port} \
            --publish={frontend_grpc_port}:{frontend_grpc_port} \
            --volume={user_dir}/.rwmeta:/esp \
            gcr.io/endpoints-release/endpoints-runtime:1 \
            --service={endpoint_service_name}.endpoints.{project}.cloud.goog \
            --rollout_strategy=managed \
            --http_port={frontend_http_port} \
            --http2_port={frontend_grpc_port} \
            --backend=grpc://{backend_host}:{backend_port} \
            --service_account_key=/esp/test-esp-service-account-creds.json
        """.format(
        user_dir=user_dir,
        backend_port=backend_port,
        frontend_grpc_port=frontend_grpc_port,
        frontend_http_port=frontend_http_port,
        backend_host=backend_host,
        project=gcloud_project,
        endpoint_service_name=endpoint_service_name
    ))


if __name__ == '__main__':
    main()



import json


class UserManagementService:
    def __init__(self, app, default_headers):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__default_headers = default_headers
        self.__options = {}

    def send_recovery_notice(self, login, app_alias, state=None):
        """
        Выслать письмо о восстановлении пароля
        """
        data = {
            "login": login,
            "appAlias": app_alias,
            "state": state,
        }
        response = self.__app.native_api_call('user-management', 'sendRecoveryNotice', data, self.__options, False, None, False, http_path="/api/meta/v1/", http_method="POST")
        return json.loads(response.text)



from metasdk.tools import exec_cmd


def run_esp_init():
    """
    Подготавливает машину разработчика к дальнейшей работе.
    Запустить один раз
    """
    exec_cmd('docker run -ti --name gcloud-config google/cloud-sdk:220.0.0-alpine gcloud auth login')


if __name__ == '__main__':
    run_esp_init()



import json
from tempfile import NamedTemporaryFile

SOURCE_FORMAT_EXTENSION = {
    'CSV': '.csv',
    'TSV': '.tsv',
    'JSON_NEWLINE': '.json'
}

class FeedService:
    def __init__(self, app, default_headers):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__default_headers = default_headers
        self.__options = {}
        self.__data_get_cache = {}
        self.__data_get_flatten_cache = {}
        self.__metadb = app.db("meta")
        self.__media = app.MediaService
        self.__starter = app.StarterService

    def get_feed(self, datasource_id):
        """
        Получение настроек для фида
        :param datasource_id: идентификатор фида
        :return: FeedDataSource
        """
        info = self.__metadb.one(
            """
            SELECT to_json(ds) as datasource
                 , to_json(fc) as connector
                 , to_json(fct) as connector_type
                 , to_json(ctp) as connector_type_preset
                 , json_build_object('email', u.email, 'full_name', u.full_name) as author_user
              FROM meta.feed_datasource ds
              LEFT JOIN meta.feed_connector fc 
                     ON fc.id=ds.connector_id
              LEFT JOIN meta.feed_connector_type fct 
                     ON fct.id=fc.connector_type_id
              LEFT JOIN meta.feed_connector_type_preset ctp 
                     ON ctp.id=ds.connector_type_preset_id
              LEFT JOIN meta.user_list u 
                     ON u.id=ds.author_user_id
             WHERE ds.id = :datasource_id::uuid
            """,
            {"datasource_id": datasource_id}
        )
        return FeedDataSource(**info)

    def get_data(self, datasource, callback):
        """
        Сохранение медиафайла
        :param task:
        :param media_metadata:
        :param file_suffix:
        :param callback:
        :return:
        """
        task = self.__app.worker.current_task
        media_metadata = datasource.connector_type_preset['preset_data']['media_metadata']
        result_data = task['result_data']
        tmp_file = NamedTemporaryFile(delete=False, suffix=SOURCE_FORMAT_EXTENSION.get(media_metadata['sourceFormat']))
        self.__app.log.info("Открываем файл", {"filename": tmp_file.name})
        with open(tmp_file.name, 'wb') as f:
            callback(f)

        self.__app.log.info("start media upload")

        result_data['stage_id'] = "persist_media_file"
        self.__starter.update_task_result_data(task)
        result = self.__media.upload(open(tmp_file.name), {
            "ttlInSec": 60 * 60 * 24,  # 24h
            "entityId": 2770,
            "objectId": task.get('data', {}).get("ds_id"),
            "info": {"metadata": media_metadata}
        })

        result_data['stage_id'] = "generate_media_finish"
        result_data['media_id'] = result['id']
        self.__starter.update_task_result_data(task)

        return result

    def datasource_process(self, datasource_id):
        """
        deprecated
        Запускает настроенные обработки в фиде
        :param datasource_id: uuid
        """
        # TODO Выпилить потом класс используется для другого
        # TODO без applicationId не выбираются поля сущностей. Подумать на сколько это НЕ нормально
        response = self.__app.native_api_call('feed', 'datasource/' + datasource_id + '/process?applicationId=1', {},
                                              self.__options, False, None, False, http_method="POST")
        return json.loads(response.text)


class FeedDataSource:
    """
    Класс хранения данных по коннектору
    """

    def __init__(self, datasource, author_user, connector, connector_type, connector_type_preset):
        self.datasource = datasource
        self.author_user = author_user
        self.connector = connector
        self.connector_type = connector_type
        self.connector_type_preset = connector_type_preset



class MailService:
    def __init__(self, app, default_headers):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__default_headers = default_headers
        self.__options = {}
        self.__data_get_cache = {}
        self.__metadb = app.db("meta")
        self.log = app.log

    def submit_mail(self, send_from, send_to, subject, body, unique_id=None):
        """
        Добавляем письмо в очередь на отправку
        :param send_from: Отправитель
        :param send_to: Получатель
        :param subject: Тема письма
        :param body: Тело письма. Можно с HTML
        :param unique_id: Уникальный идентификатор письма. Обычно что-то вроде md5 + человекочитаемый префикс подходят лучше всего. Письмо с одинаковым unique_id не будет добавлено
        """
        self.__metadb.update("""
        INSERT INTO meta.mail("template", "from", "to", "subject", "body", "attachments", "unique_id")
        VALUES ('meta', :send_from, :send_to, :subject, :body, null, :unique_id)
        ON CONFLICT (unique_id) DO NOTHING
        """, {
            "send_from": send_from,
            "send_to": send_to,
            "subject": subject,
            "body": body,
            "unique_id": unique_id
        })



import logging

from metasdk.logger import LOGGER_ENTITY


class Logger:
    """
    Прослойка для упрощения апи логгера
    """

    def set_entity(self, key, value):
        if value is None:
            self.remove_entity(key)
        else:
            LOGGER_ENTITY[key] = value

    def remove_entity(self, key):
        LOGGER_ENTITY.pop(key, None)

    def info(self, msg, context=None):
        if context is None:
            context = {}
        logging.info(msg, extra={'context': context})

    def warning(self, msg, context=None):
        if context is None:
            context = {}
        logging.warning(msg, extra={'context': context})

    def error(self, msg, context=None):
        if context is None:
            context = {}
        logging.error(msg, extra={'context': context})

    def critical(self, msg, context=None):
        if context is None:
            context = {}
        logging.critical(msg, extra={'context': context})

    def exception(self, msg, context=None):
        if context is None:
            context = {}
        logging.exception(msg, extra={'context': context})
